package ru.landgrafhomyak.serdha.api.v0.queries

import ru.landgrafhomyak.serdha.api.v0.misc.Column
import ru.landgrafhomyak.serdha.api.v0.table.Table
import ru.landgrafhomyak.serdha.api.v0.misc.DatabaseType
import ru.landgrafhomyak.serdha.api.v0.misc.DatabaseTypesProvider
import ru.landgrafhomyak.serdha.api.v0.misc.QueryParam
import ru.landgrafhomyak.serdha.api.v0.runtime.ParametersSetter
import ru.landgrafhomyak.serdha.api.v0.table.UniqueIndex

public interface InsertQuery<QueryUserExtension : Any> {
	public val userExtension: QueryUserExtension

	@Suppress("ClassName")
	public interface _UpsertClauseScope<QueryUserExtension : Any, TargetTableUserExtension : Any> {
		public fun keepColumn(c: Column<*, *, TargetTableUserExtension>)
		public fun overrideColumn(c: Column<*, *, TargetTableUserExtension>)
		public fun setConstant(c: QueryParam<*, *, QueryUserExtension>)
	}

	@Suppress("ClassName")
	public interface _CommonInsertCreatorScope<QueryUserExtension : Any, TargetTableUserExtension : Any> {
		public val types: DatabaseTypesProvider

		public fun onConflictUpdate(u: UniqueIndex<TargetTableUserExtension>, c: (_UpsertClauseScope<QueryUserExtension, TargetTableUserExtension>) -> Unit)
		public fun onConflictThrow(u: UniqueIndex<TargetTableUserExtension>)
		public fun onConflictSkip(u: UniqueIndex<TargetTableUserExtension>)
	}


	public interface SingleRow<QueryUserExtension : Any, TargetTableUserExtension : Any> : InsertQuery<QueryUserExtension> {
		public interface Creator<QueryUserExtension : Any, TargetTableUserExtension : Any> {
			public fun createInsertSingleRowWithReturning(table: Table<TargetTableUserExtension, *>, creator: Scope<QueryUserExtension, TargetTableUserExtension>): QueryUserExtension

			public interface Scope<QueryUserExtension : Any, TargetTableUserExtension : Any> : _CommonInsertCreatorScope<QueryUserExtension, TargetTableUserExtension> {
				public fun <RT, DT : DatabaseType<RT>> queryParam(name: String, type: DT): QueryParam<RT, DT, QueryUserExtension>

				public fun <RT, DT : DatabaseType<RT>> insertConstant(column: Column<RT, DT, TargetTableUserExtension>, paramName: String = column.name): QueryParam<RT, DT, QueryUserExtension>
				public fun <RT, DT : DatabaseType<RT>> insertConstant(column: Column<RT, DT, TargetTableUserExtension>, param: QueryParam<RT, DT, QueryUserExtension>)

				public fun onConflictReturnOnly(u: UniqueIndex<TargetTableUserExtension>)

				public fun <RT, DT : DatabaseType<RT>> returnInsertedValue(c: Column<RT, DT, TargetTableUserExtension>): Column<RT, DT, QueryUserExtension>
				public fun <RT, DT : DatabaseType<RT>> returnParam(c: QueryParam<RT, DT, TargetTableUserExtension>): Column<RT, DT, QueryUserExtension>
			}
		}
	}


	public interface FromSubquery<QueryUserExtension : Any, TargetTableUserExtension : Any> : InsertQuery<QueryUserExtension> {
		public interface Creator<QueryUserExtension : Any, TargetTableUserExtension : Any, SourceQueryUserExtension : Any> {
			public fun createInsertSingleRowWithReturning(table: Table<TargetTableUserExtension, *>, creator: Scope<QueryUserExtension, TargetTableUserExtension, SourceQueryUserExtension>): QueryUserExtension

			public interface Scope<QueryUserExtension : Any, TargetTableUserExtension : Any, SourceQueryUserExtension : Any> : _CommonInsertCreatorScope<QueryUserExtension, TargetTableUserExtension> {
				public fun <RT, DT : DatabaseType<RT>> queryParam(name: String, type: DT): QueryParam<RT, DT, QueryUserExtension>

				public val sourceQueryParams: ParametersSetter.WithRedirect<SourceQueryUserExtension, QueryUserExtension>

				public fun <RT, DT : DatabaseType<RT>> insertConstant(column: Column<RT, DT, TargetTableUserExtension>, paramName: String = column.name): QueryParam<RT, DT, QueryUserExtension>
				public fun <RT, DT : DatabaseType<RT>> insertConstant(column: Column<RT, DT, TargetTableUserExtension>, param: QueryParam<RT, DT, QueryUserExtension>)
				public fun <RT, DT : DatabaseType<RT>> insertFromSubquery(column: Column<RT, DT, TargetTableUserExtension>, subquery: QueryParam<RT, DT, SourceQueryUserExtension>)

				public fun onConflictReturnOnly(u: UniqueIndex<TargetTableUserExtension>)

				public fun <RT, DT : DatabaseType<RT>> returnFromSubquery(c: Column<RT, DT, SourceQueryUserExtension>): Column<RT, DT, QueryUserExtension>
				public fun <RT, DT : DatabaseType<RT>> returnNewValue(c: Column<RT, DT, TargetTableUserExtension>): Column<RT, DT, QueryUserExtension>
				public fun <RT, DT : DatabaseType<RT>> returnParam(c: QueryParam<RT, DT, TargetTableUserExtension>): Column<RT, DT, QueryUserExtension>
			}
		}
	}

	public interface MultipleRows<QueryUserExtension : Any, TargetTableUserExtension : Any> : InsertQuery<QueryUserExtension> {
		public interface Creator<QueryUserExtension : Any, TargetTableUserExtension : Any> {
			public fun createInsertMultipleRows(table: Table<TargetTableUserExtension, *>, creator: Scope<QueryUserExtension, TargetTableUserExtension>): QueryUserExtension

			public interface Scope<QueryUserExtension : Any, TargetTableUserExtension : Any> : _CommonInsertCreatorScope<QueryUserExtension, TargetTableUserExtension> {
				public fun <RT, DT : DatabaseType<RT>> inputColumn(name: String, type: DT): QueryParam<RT, DT, QueryUserExtension>

				public fun <RT, DT : DatabaseType<RT>> insertColumn(dst: Column<RT, DT, TargetTableUserExtension>, name: String = dst.name)
				public fun <RT, DT : DatabaseType<RT>> insertColumn(dst: Column<RT, DT, TargetTableUserExtension>, src: QueryParam<RT, DT, QueryUserExtension>)
			}
		}
	}
}