package ru.landgrafhomyak.serdha.api.v0.table

import kotlin.jvm.JvmName
import ru.landgrafhomyak.serdha.api.v0.misc.Column
import ru.landgrafhomyak.serdha.api.v0.misc.Expression
import ru.landgrafhomyak.serdha.api.v0.misc.DatabaseType

/**
 * Descriptor of default value attached to [column][Column]. Used for schema manipulations.
 *
 * @param RT Type in programming language to which database type is converted.
 * @param DT Descriptor of column type on database side.
 * @param TableUserExtension Type of [owner's table][Column.table] user expression for static reporting errors when this descriptor passed to wrong table.
 */
public interface DefaultConstraint<RT, DT : DatabaseType<RT>, TableUserExtension : Any> {
	/**
	 * Name of constraint for debugging, errors and raw schema access.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("name")
	public val name: String

	/**
	 * Table, that contains this constraint. For debugging.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("table")
	public val table: Table<TableUserExtension, *>

	/**
	 * Column which will get value from this constraint if not initialized manually. For debugging.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("attachedToColumn")
	public val attachedToColumn: Column<RT, DT, TableUserExtension>

	/**
	 * Expression used to initialize cell.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("expression")
	public val expression: Expression<RT, DT, TableUserExtension>
}